/**
 * @(#)RobotController.java 1.1 2002/10/12
 *
 * Copyright Wright State University. All Rights Reserved.
 *
 * This file is part of the WSU Khepera Simulator.
 *
 * This file may be distributed under the terms of the KSIM PUBLIC LICENSE 1.0
 * (KPL) appearing in the file WSU_Khepera_Sim_license.txt included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * For information on the KSIM PUBLIC LICENSE 1.0 see:
 * 		http://gozer.cs.wright.edu/ksim/license.html
 *
 * For information on the WSU Khepera Simulator see:
 * 		http://gozer.cs.wright.edu and follow the links.
 *
 * Contact robostaff@gozer.cs.wright.edu if any conditions of this licensing are
 * not clear to you.
 */

package edu.wsu.KepheraSimulator;


/**
 * A <code>RobotController</code> is an abstract class that should be extended
 * by any controller written for the WSU Kephera Simulator.  This class provides
 * the interface to the simulator and contains the threading construct used
 * to execute the controller.
 */
public abstract class RobotController implements Controller{

  /**
   * Time interval in milliseconds to wait until the next invocation of the
   * method <code>doWork</code>.  The wait time does not begin until
   * <code>doWork</code> has returned.  The default value is 5ms.
   */
  private long waitTime = 5;

  /**
   * Internal flag which tells this object if it should terminate or not.
   * This varible is initialized to <tt>false</tt>.  When set to <tt>true</tt>
   * this object will call the <code>close</code> method.  Next, this executing
   * thread will die by returning from the <code>run</code> method.
   */
  private boolean finished = false;

  /**
   * Internal conditional variable which represents the current running state
   * of this controller.
   */
  private boolean running = false;

  /**
   * <tt>Thread</tt> object for which this <tt>Runnable</tt> instance of
   * <tt>JTPProgram</tt> will be created.  It is important to control all of
   * the thread behavior with this object.
   */
  private Thread thread = null;

  /**
   * The name of this controller.
   */
  private String name = null;

  /**
   * A reference to the CurrentRobotState object.  This reference will be used
   * to return the current robot state at any time.
   */
  private CurrentRobotState currentRobotState = null;
  private Motor motor = null;
  private Sensor[] sensors = null;

	private KSFrame ksf;		//NEW

  /**
   * Allocates a new <code>RobotController</code> object.
   */
  public RobotController() { }

  /**
   * This method should only be used by the simulator.  This method is used
   * by the simulator to initialize this controller with references to the
   * local simulator.  Also a <code>Thread</code> object is created to execute
   * this <code>RobotCcontroller</code>.
   * @param name the name of this controller
   * @param currentRobotState a reference to the current robot state
   */
  protected void initialize(String name, CurrentRobotState currentRobotState, KSFrame ksf) {
    if (!isRunning()) {
      this.name = name;         // set this programs name
      this.currentRobotState = currentRobotState;
      this.ksf = ksf;											//NEW
      motor = currentRobotState.getMotorState();
      sensors = currentRobotState.getSensorValues();
      thread = new Thread(this, name);  // new thread for controller
      thread.start();           // begin the execution of this controller
    }
  } // initialize

  /**
   * This method is used by the controller to set the delay before the next
   * execution of the doWork method.  The delay is usually set in the
   * controllers constructor.  After the doWork method finishes the
   * execution thread will wait for the specified amount of time to elapse
   * before executing doWork again.  The default delay time is 5 milliseconds.
   * @param waitTime the wait time in milliseconds
   */
  protected void setWaitTime(long waitTime) {
    this.waitTime = waitTime;
  }

  /**
   * This method is used by the simulator in order to tell this controller
   * to stop running.  This allows for a graceful termination of this
   * controller thread.
   * @param finished set this flag <tt>true</tt> to terminate the controller
   */
  protected void setFinished(boolean finished) {
    this.finished = finished;
  }

  /**
   * This method is used by the simulator to check if the controller is
   * currently running.  In this context running means the controller thread
   * could be executing the doWork method or in between doWork calls.
   * @return <tt>true</tt> if this controller is running
   */
  protected boolean isRunning() {
    return running;
  }

  /**
   * This method should not be called by the controller
   * @see java.lang.Runnable
   */
  public void run() {
    running = true;
    try {
      while (!finished) {
        doWork();
        sleep(waitTime);
      }
      close();
    } catch (Exception e) { /* log exceptions */ }
    running = false;
  } // run

  /**
   * This method should be used by the controller if an extra delay is needed
   * at some point in the controllers doWork method.  It is very important that
   * this method be used to insert delays rather than the objects
   * <code>wait</code> method, otherwise there could be drastic performance
   * problems.
   * @param timeout the delay time in milliseconds
   */
  public void sleep(long timeout) {
    try {
      thread.sleep(timeout);
    } catch (Exception e) { }
  }

  //NEW
  public void setMsg(String s){
	  ksf.setMsg(s);
	  System.out.println("In RobotController.setMsg()");
  }//NEW

  //****** Robot State accessor methods ******/

  /**
   * This method is provided so the controller can control the motor speeds of
   * the left and right wheels.  The valid range for motor speeds are between
   * the range of 10 and -10, not inclusive.
   * @param left the left wheel speed
   * @param right the right wheel speed
   */
  public void setMotorSpeeds(int left, int right) {
    motor.setMotorSpeeds(left, right);
  }

  /**
   * This method is provided so the controller can control the left motor speed.
   * The valid range for motor speeds are between the range of 10 and -10,
   * not inclusive.
   * @param speed the left wheel speed
   */
  public void setLeftMotorSpeed(int speed) {
    motor.setLeftMotorSpeed(speed);
  }

  /**
   * This method is provided so the controller can control the right motor speed.
   * The valid range for motor speeds are between the range of 10 and -10,
   * not inclusive.
   * @param speed the right wheel speed
   */
  public void setRightMotorSpeed(int speed) {
    motor.setRightMotorSpeed(speed);
  }

  /**
   * This method is used to attain the current position of the right wheel. On
   * the actual robot, this information is generated by incrementing/decrementing
   * a counter based on shaft encoder readings.
   * @return right motor position
   */
  public long getRightWheelPosition() {
    return motor.getRightPosition();
  }

  /**
   * This method is used to attain the current position of the left wheel. On
   * the actual robot, this information is generated by incrementing/decrementing
   * a counter based on shaft encoder readings.
   * @return left motor position
   */
  public long getLeftWheelPosition() {
    return motor.getLeftPosition();
  }


  /**
   * This method is provided so the controller can check the current state of
   * the robot gripper.  The valid states for the gripper are: KSGripperStates.GRIP_OPEN
   * and KSGripperStates.GRIP_CLOSED.
   * @return the gripper state
   */
  public int getGripperState() {
    return currentRobotState.getGripperState();
  }

  /**
   * This method is provided so the controller can set the gripper state of the
   * robot.  The valid states are: KSGripperStates.GRIP_OPEN and KSGripperStates.GRIP_CLOSED.
   * @param state the gripper state
   */
  public void setGripperState(int state) {
    sleep(100);  // added 8/6/03 - sp - simulates time delay on real robot
    currentRobotState.setGripperState(state);
  }

  /**
   * This method is provided so the controller can check the current state of
   * the robot arm.  The valid states for the arm are:
   *    KSGripperStates.ARM_UP and
   *    KSGripperStates.ARM_DOWN
   * @return the arm state
   */
  public int getArmState() {
    return currentRobotState.getArmState();
  }

  /**
   * This method is provided so the controller can set the current state of
   * the robot arm.  The valid states for the arm are:
   *    KSGripperStates.ARM_UP and
   *    KSGripperStates.ARM_DOWN
   * @param state the arm state
   */
  public void setArmState(int state) {
    sleep(100);  // added 8/6/03 - sp - simulates time delay on real robot
    currentRobotState.setArmState(state);
  }

  /**
   * This method is provided so the controller can check if an object is
   * present inside the gripper.
   * @return <tt>true</tt> if an object is present, <tt>false</tt> otherwise
   */
  public boolean isObjectPresent() {
    sleep(100);  // added 8/6/03 - sp - may not need
    return currentRobotState.isObjectPresent();
  }

  /**
   * This method is provided so the controller can check if an object is
   * currently held by the gripper.
   * @return <tt>true</tt> if an object is held, <tt>false</tt> otherwise
   */
  public boolean isObjectHeld() {
    return currentRobotState.isObjectHeld();
  }

  /**
   * This method is provided so the controller can get the light value of a
   * specific sensor given the sensor ID.  The sensor ID's range from 0 to 7.
   * TODO: describe where each sensor is located on the robot.
   * TODO: describe the range of light values and meaning of the scale.
   * @return the light value
   */
  public int getLightValue(int sensorID) {
    return sensors[sensorID].getLightValue();
  }

  /**
   * This method is provided so the controller can get the distance value of a
   * specific sensor given the sensor ID.  The sensor ID's range from 0 to 7.
   * TODO: describe where each sensor is located on the robot.
   * TODO: describe the range of distance values and meaning of the scale.
   * @return the distance value
   */
  public int getDistanceValue(int sensorID) {
    return sensors[sensorID].getDistValue();
  }

  /**
   * This method is provided so the controller can get the resistivity of the
   * current object between the gripper.  This feature has not yet been
   * implemented.
   * @return the resistivity value
   */
  public int getResistivity() {
    return currentRobotState.getResistivity();
  }

} // RobotController